#!/usr/bin/python

import collections

COLOURS = ["white", "bisque3", "brown", "chartreuse", "coral",
    "cyan3", "darkorchid1", "gold", "burlywood1",
    "navajowhite1", "oldlace", "hotpink"]
RUN_MASK = 0xffffffff

class Method:
    def __init__(self, start, name, size, call_count, region):
        self.start = start
        self.name = name
        self.size = size
        self.call_count = call_count
        self.region = region
        self.number = -1
        self.cycle_children = []
        self.tree_children = []
        self.parent = None
        self.previous_clone = None
        self.next_clone = None
        self.original = self
        self.depth = None
        self.clone_count = 0
        self.split_count = 0
        self.loop_from = []
        self.loop_to = []
        self.sequence = []
        self.map_to = None

    def clone(self, new_parent):
        self.original.clone_count += 1
        other = Method(self.start, 
                        "%s_c%u" % (self.original.name,
                                    self.original.clone_count), 
                        self.size, self.call_count, self.region)
        other.previous_clone = self
        self.next_clone = other
        other.original = self.original
        other.parent = new_parent
        other.depth = self.depth
        other.tree_children = self.tree_children[:]
        other.cycle_children = self.cycle_children[:]
        other.map_to = self.map_to
        return other

    def split(self, max_size):
        assert self.size > max_size
        self.split_count += 1
        self.size -= max_size
        other = Method(self.start, 
                        "%s_s%u" % (self.name, self.split_count), 
                        max_size, self.call_count, self.region)
        other.tree_children = self.tree_children
        other.cycle_children = self.cycle_children
        self.tree_children = [(other, 1, True)]
        self.cycle_children = []
        other.depth = self.depth
        other.parent = self
        return other

class Graph:
    def __init__(self):
        self.__methods = dict()
        self.__edges = []
        self.__loops = []
        self.__sequences = []

        self.methods = []

        self.root = None

    def register_method(self, start, name, size, call_count, region = 0):
        assert len(name) != 0

        size = max(size, 4)
        m = self.__methods.get(start, None)
        if m == None:
            m = Method(start, name, size, call_count, region)
            self.__methods[start] = m
        else:
            assert m.name == name
            m.size = max(m.size, size)
            m.call_count += call_count

        return m

    def register_link(self, parent_start, child_start, link_count):
        self.__edges.append((parent_start, child_start, link_count))

    def register_loop(self, header_start, footer_start, link_count):
        self.__loops.append((header_start, footer_start, link_count))

    def register_meth_sequence(self, parent_start, child_items):
        self.__sequences.append((parent_start, child_items))

    def read_python(self, graph_in):
        area = {
            "method" : self.register_method,
            "link" : self.register_link,
            "loop" : self.register_loop,
            "meth_sequence" : self.register_meth_sequence,
            "block" : lambda size: ('B', size),
            "call" : lambda address: ('C', address),
            }
        execfile(graph_in, area, area)
        assert len(self.__methods) != 0

    def finalise(self, use_sequences = False):
        # Make list of methods
        assert len(self.methods) == 0
        assert self.root == None

        for (_, parent) in self.__methods.iteritems():
            if parent.name == "main":
                self.root = parent

            parent.depth = None
            self.methods.append(parent)

        # Initial sort by starting address
        self.methods.sort(key = lambda parent: parent.start)

        # De-reference edges
        edges = []
        if not use_sequences:
            for (parent_ref, child_ref, cost) in self.__edges:
                edges.append((self.__methods[parent_ref],
                            self.__methods[child_ref], cost))

        self.__loops.sort()
        self.__loops.reverse()
        back_edges = []
        for (header_ref, footer_ref, cost) in self.__loops:
            back_edges.append((self.__methods[header_ref],
                        self.__methods[footer_ref], cost))

        for (parent_ref, child_items_ref) in self.__sequences:
            parent = self.__methods[parent_ref]
            child_items = []
            for (mode, ref) in child_items_ref:
                if mode == 'B':
                    # block
                    child_items.append((mode, ref))
                elif mode == 'C':
                    # call
                    child = self.__methods[ref]
                    child_items.append((mode, child))
                    if use_sequences:
                        edges.append((parent, child, parent.call_count))
                else:
                    assert False

            parent.sequence = child_items

        # Build parent->child data structure
        children = dict()
        if use_sequences:
            for parent in self.methods:
                children[parent] = []

            for (parent, child, cost) in edges:
                c = children[parent]
                c.append((child, cost))
                parent.depth = 0

        else:
            for parent in self.methods:
                children[parent] = dict()

            for (parent, child, cost) in edges:
                c = children[parent]
                c[child] = c.get(child, 0) + cost

            for parent in self.methods:
                children[parent] = children[parent].items()
                parent.depth = 0

        # Label depths
        reached_this_time = set()
        def depth_assign(parent, depth):

            if parent in reached_this_time:
                assert parent.depth < depth
                return # cycle detect

            parent.depth = depth = max(parent.depth, depth)
            reached_this_time.add(parent)

            for (child, cost) in children[parent]:
                depth_assign(child, depth + 1)

            assert parent in reached_this_time
            reached_this_time.discard(parent)

        depth_assign(self.root, 0)

        # Assignment of children
        for parent in self.methods:
            for (child, cost) in children[parent]:
                if child.depth > parent.depth:
                    parent.tree_children.append((child, cost))
                else:
                    parent.cycle_children.append((child, cost))

        # Assignment of back_edges
        for (header, footer, cost) in back_edges:
            header.loop_from.append((footer, cost))
            footer.loop_to.append((header, cost))

        self.__sort()

    def __sort(self):
        # sort by depth
        self.methods.sort(key = lambda parent: parent.depth)
        assert self.methods[0] == self.root
        assert self.root.depth == 0

        # renumber methods
        for (i, parent) in enumerate(self.methods):
            assert parent.depth != None
            parent.number = i
            parent.parent = None

        for parent in self.methods:
            for (child, _) in parent.tree_children:
                child.parent = parent
                
        for parent in self.methods:
            assert (parent.parent != None) or (parent == self.root)

    def write_python(self, graph_out):
        f = file(graph_out, "wt")
        for parent in self.methods:
            f.write("method(0x%x, '%s', 0x%x, 0x%x, %u)\n" % (
                parent.start, parent.name, parent.size, 
                parent.call_count, parent.region))

        for parent in self.methods:
            for (child, cost) in parent.tree_children:
                f.write("link(0x%x, 0x%x, 0x%x)\n" % (
                        parent.start, child.start, cost))

        for header in self.methods:
            for (footer, cost) in header.loop_from:
                f.write('loop(0x%x, 0x%x, 0x%x)\n' % (
                        header.start, footer.start, cost))

        f.close()

    def write_graphviz(self, graph_out, custom_method_label = None):
        out = file(graph_out, "wt")
        out.write("digraph G {")

        region_colour = {0:0}
        num_colours = len(COLOURS)
        subtree_root = set()

        for parent in self.methods:
            parent_colour = region_colour[parent.region]
            choose_colour = (parent_colour + 3) % num_colours

            for (child, _) in parent.tree_children:
                if child.region != parent.region:
                    region_colour[child.region] = child_colour = choose_colour
                    subtree_root.add(child.number)

                    while choose_colour in (child_colour, parent_colour):
                        choose_colour = (choose_colour + 1) % num_colours

        for parent in self.methods:
            shape = "box"
            #if (parent == self.root) or (parent.number in subtree_root):
            #    shape = "ellipse"

            p = { "nodenumber" : parent.number,
                    "shape" : shape,
                    "name" : parent.name,
                    "size" : parent.size,
                    "colour" : COLOURS[region_colour[parent.region]], 
                }

            if custom_method_label == None:
                label = "(%(nodenumber)u) %(name)s %(size)u" % p
            else:
                label = custom_method_label(parent)

            p["label"] = label
            out.write("""
node [shape=%(shape)s, style=filled, color=black, fillcolor=%(colour)s,
        label="%(label)s"] N%(nodenumber)u;""" % p)


        out.write("\n")

        for parent in self.methods:
            for (child, cost) in parent.tree_children:
                label = "%u" % cost
                out.write('N%u -> N%u [label="%s"];\n' % (
                        parent.number, child.number, label))

        for parent in self.methods:
            for (child, cost) in parent.cycle_children:
                label = "%u C" % cost
                out.write('N%u -> N%u [label="%s"];\n' % (
                        parent.number, child.number, label))

        for header in self.methods:
            for (footer, cost) in header.loop_from:
                out.write('N%u -> N%u [label="%u L!"];\n' % (
                        footer.number, header.number, cost))

        out.write("}\n\n")

    def is_tree(self):
        # Two separate checks for tree properties
        # First check:
        for parent in self.methods:
            for (child, _) in parent.tree_children:
                if child.parent != parent:
                    return False

            if parent.parent == None:
                if parent != self.root:
                    return False

        # Second check:
        reached_once = set()
        def is_tree_r(root):
            if root in reached_once:
                return False # contains multiple paths to the same node

            reached_once.add(root)
            for (child, _) in root.tree_children:
                if not is_tree_r(child):
                    return False

            return True

        return is_tree_r(self.root)


    def sequencise(self, spm_size):
        assert self.is_tree()

        new_graph = Graph()

        def process(parent, return_to, return_count):
            start = parent.start
            current = start
            previous = None

            last_method = None
            child_nr = 0

            for (mode, ref) in parent.sequence:
                if mode == 'B':
                    size = ref
                    real_end = current + size
                    division = 1
                    while (((size / division) + 3) & ~3) > spm_size:
                        division += 1

                    size = (((size / division) + 3) & ~3)
                    assert size > 0

                    for i in xrange(division):
                        name = '%s_%x_%x' % (parent.name, current, size)
                        if (parent.number == 0) and (current == parent.start):
                            name = "main"

                        last_method = new_graph.register_method(
                                start = current, 
                                name = name,
                                size = min(size, real_end - current),
                                call_count = parent.call_count)
                        last_method.map_to = parent

                        if previous != None:
                            new_graph.register_link(previous, 
                                    current, parent.call_count)

                        previous = current
                        current += size

                    current = real_end

                elif mode == 'C':
                    new_graph.register_link(current, 
                            ref.start, parent.call_count)
                    assert child_nr < len(parent.tree_children)
                    (child, cc) = parent.tree_children[child_nr]
                    print parent.name, child_nr, child.name
                    assert child.start == ref.start, ref.name
                    process(child, current, cc)
                    child_nr += 1

                else:
                    assert False

            if ((return_to != None) and (last_method != None)):
                new_graph.register_loop(return_to, last_method.start,
                        return_count)

        process(self.root, None, 0)
        new_graph.finalise()
        return new_graph

    def treeise(self):
        clones_added = True
        while clones_added:
            clones = []
            for parent in self.methods:
                old_tc = parent.tree_children
                parent.tree_children = new_tc = []
                for (child, cost) in old_tc:
                    assert child.depth > parent.depth

                    if child.parent == parent:
                        # No cloning required
                        new_tc.append((child, cost))
                    else:
                        # Make a clone of this subtree
                        new_child = child.clone(parent)
                        new_tc.append((new_child, cost))
                        clones.append(new_child)

            clones_added = False
            if len(clones) != 0:
                clones_added = True
                self.methods.extend(clones)

        self.__sort()
        assert self.is_tree()

    def get_min_spm_size(self):
        size = 0
        for parent in self.methods:
            size = max(parent.size, size)
        return size

    def write_csearch(self, graph_out, spm_size,
                    mode = "C", quantum = 4, 
                    fixed_cost = 0):
        assert self.is_tree()
        f = file(graph_out, "wt")
        f.write("%u %u %s %u %u\n" % (
                spm_size, len(self.methods), mode, quantum, fixed_cost))

        for child in self.methods:
            p = child.parent
            if p == None:
                assert child == self.root
                pn = 0
                entry = 1
            else:
                pn = p.number
                entry = None
                for (c, cost) in p.tree_children:
                    if c == child:
                        entry = cost

                assert entry != None
                assert pn < child.number

            cost = entry 
            assert cost > 0

            f.write("%-8u %-8u %-8u %u\n" % (
                child.size, pn, cost, child.region))

        if mode in 'wW':
            count = 0
            for header in self.methods:
                count += len(header.loop_from)

            f.write("%u\n" % count)
            for header in self.methods:
                for (footer, cost) in header.loop_from:
                    f.write("%-4u %-4u %u\n" % (
                            header.number, footer.number, cost))

        f.write("0\n") # no value
        f.close()

    def read_csearch(self, graph_in):
        fd = file(graph_in, "rt")

        # header
        fields = fd.readline().split()
        assert len(fields) == 5
        spm_size = int(fields[0])
        assert len(self.methods) == int(fields[1])
        mode = fields[2]
        quantum = int(fields[3])
        fixed_cost = int(fields[4])

        # methods
        for node_number in xrange(len(self.methods)):
            fields = fd.readline().split()
            assert len(fields) == 4
            assert self.methods[node_number].size == int(fields[0]), (node_number,
                    self.methods[node_number].size, fields)
            p = int(fields[1]) 
            if node_number == 0:
                assert p == 0
                p = None
            else:
                assert self.methods[node_number].parent.number == p

            self.methods[node_number].region = int(fields[3])

        # back edges
        if mode in "wW":
            fields = fd.readline().split()
            assert len(fields) == 1
            p = int(fields[0])
            count = 0
            for header in self.methods:
                count += len(header.loop_from)

            assert count == p

            for header in self.methods:
                for (footer, cost) in header.loop_from:
                    fields = fd.readline().split()
                    assert len(fields) == 3
                    assert header == self.methods[int(fields[0])]
                    assert footer == self.methods[int(fields[1])]
                    assert cost == int(fields[2])

        fields = fd.readline().split()
        assert len(fields) == 1
        value = int(fields[0])
        fd.close()

        return {
                "value" : value,
                "spm_size" : spm_size, 
                "mode" : mode, 
                "quantum" : quantum, 
                "fixed_cost" : fixed_cost,
                }

    def split_large_methods(self, bigger_than):
        for i in xrange(len(self.methods)):
            m1 = self.methods[i]
            while m1.size > bigger_than:
                print 'split', m1.name
                m2 = m1.split(bigger_than)
                self.methods.append(m2)

        changing = True
        while changing:
            changing = False
            for parent in self.methods:
                for (child, cost) in parent.tree_children:
                    if child.depth <= parent.depth:
                        child.depth = parent.depth + 1
                        changing = True
        
        self.__sort()

    def remove_leaf(self):
        leaves = 0
        leaf = None
        for m in self.methods:
            if (len(m.tree_children) == 0) and (len(m.loop_to) == 0):
                leaves += 1
                leaf = m
                print 'leaf is: (%u) 0x%x' % (leaf.number, leaf.start)

        assert leaves == 1, leaves
        assert leaf != None
        assert leaf.size == 4

        assert self.methods[leaf.number] == leaf
        assert leaf.parent != None
        assert len(leaf.loop_to) == 0
        assert len(leaf.tree_children) == 0

        nc = []
        for (child, cost) in leaf.parent.tree_children:
            if child != leaf:
                nc.append((child, cost))

        leaf.parent.tree_children = nc
        self.methods.pop(leaf.number)
        self.__sort()

    def get_region_sizes(self):
        region_contents = dict()
        for m in self.methods:
            rc = region_contents.get(m.region, None)
            if rc == None:
                region_contents[m.region] = rc = set()
                
            for i in xrange(m.start, m.start + m.size, 1):
                rc.add(i & RUN_MASK)

        no_overlap_region_size = dict()
        for m in self.methods:
            no_overlap_region_size[m.region] = (m.size +
                    no_overlap_region_size.get(m.region, 0))

        region_size = dict()
        for (number, rc) in region_contents.iteritems():
            region_size[number] = len(rc)
            assert region_size[number] <= no_overlap_region_size[number]

        return region_size

    def get_value(self, fixed_cost):
        region_size = self.get_region_sizes()

        total = 0
        for mx in self.methods:
            for (my, cost) in mx.tree_children:
                if mx.region != my.region:
                    total += (region_size[my.region] + fixed_cost) * cost

            for (my, cost) in mx.loop_from:
                if mx.region != my.region:
                    total += (region_size[mx.region] + fixed_cost) * cost

        total += region_size[self.methods[0].region] + fixed_cost
        assert total > 0

        return total


